// $Id: helptopic.cc,v 1.6 2003/05/28 15:56:50 toast Exp $

#include <fstream>
#include <string>
#include <list>
#include <algorithm>

#include "helptopic.h"
#include "config.h"
#include "util.h"

HelpTopic::HelpTopic(void)
{
  topics.clear();
  syntax.clear();
  description.clear();
  example.clear();
  helpLink.clear();
  subTopic.clear();
  flags = UF_NONE;
  error = 0;
}


HelpTopic::HelpTopic(const HelpTopic & copy)
{
  topics = copy.topics;
  syntax = copy.syntax;
  description = copy.description;
  example = copy.example;
  helpLink = copy.helpLink;
  subTopic = copy.subTopic;
  flags = copy.flags;
  error = 0;
}


HelpTopic::~HelpTopic(void)
{
}


bool
HelpTopic::isTopic(const std::string & topic)
{
  StrVector words;
  int count = StrSplit(words, topic, " ", true);

  std::string topicMaj, topicMin;

  if (count > 0)
  {
    topicMaj = words[0];
    if (count > 1)
    {
      topicMin = words[1];
    }
  }

  for (StrVector::iterator pos = topics.begin();
    pos != topics.end(); ++pos)
  {
    std::string current = *pos;

    if ((topicMaj == FirstWord(current)) && (topicMin == FirstWord(current)))
    {
      return true;
    }
  }

  return false;
}


StrList
HelpTopic::getHelp(void)
{
  StrList result;

  result.clear();

  if (topics.size() <= 0)
  {
    return result;
  }

  if (syntax.size() > 0)
  {
    result.push_back("\002Syntax:\002");
    for (StrList::iterator pos = syntax.begin();
      pos != syntax.end(); ++pos)
    {
      result.push_back(std::string("  .") + *pos);
    }
  }

  if (description.size() > 0)
  {
    result.push_back("\002Description:\002");
    for (StrList::iterator pos = description.begin();
      pos != description.end(); ++pos)
    {
      result.push_back(std::string("  ") + *pos);
    }
  }

  if (example.size() > 0)
  {
    result.push_back("\002Example:\002");
    for (StrList::iterator pos = example.begin();
      pos != example.end(); ++pos)
    {
      result.push_back(std::string("  ") + *pos);
    }
  }

  result.push_back("\002Required Flags:\002");
  if (flags == UF_NONE)
  {
    result.push_back("  None");
  }
  else
  {
    if (flags & UF_CHANOP)
    {
      result.push_back("  C - Chanop");
    }
    if (flags & UF_DLINE)
    {
      result.push_back("  D - Dline");
    }
    if (flags & UF_GLINE)
    {
      result.push_back("  G - Gline");
    }
    if (flags & UF_KLINE)
    {
      result.push_back("  K - Kline");
    }
    if (flags & UF_MASTER)
    {
      result.push_back("  M - Master");
    }
    if (flags & UF_OPER)
    {
      result.push_back("  O - Oper");
    }
    if (flags & UF_REMOTE)
    {
      result.push_back("  R - Remote");
    }
    if (flags & UF_WALLOPS)
    {
      result.push_back("  W - Wallops");
    }
  }

  if (helpLink.size() > 0)
  {
    std::string row = std::string("");

    result.push_back("\002See Also:\002");
    for (StrList::iterator pos = helpLink.begin();
      pos != helpLink.end(); ++pos)
    {
      if (row.length() == 0)
      {
        row = DownCase(*pos);
      }
      else
      {
        if ((row.length() + pos->length() + 2) > 40)
        {
          result.push_back(std::string("  ") + row);
	  row = DownCase(*pos);
	}
	else
	{
	  row += std::string("  ") + DownCase(*pos);
	}
      }
    }
    if (row.length() > 0)
    {
      result.push_back(std::string("  ") + row);
    }
  }

  if (subTopic.size() > 0)
  {
    std::string row = std::string("");

    result.push_back("\002Sub-Topics:\002 (.help set <sub-topic>)");
    for (StrList::iterator pos = subTopic.begin();
      pos != subTopic.end(); ++pos)
    {
      if (row.length() == 0)
      {
        row = DownCase(*pos);
      }
      else
      {
        if ((row.length() + pos->length() + 2) > 40)
	{
          result.push_back(std::string("  ") + row);
	  row = DownCase(*pos);
	}
	else
	{
	  row += std::string("  ") + DownCase(*pos);
	}
      }
    }
    if (row.length() > 0)
    {
      result.push_back(std::string("  ") + row);
    }
  }

  return result;
}


StrList
HelpTopic::getHelp(const std::string & topic)
{
  std::string	line;
  bool		inTopic = false;

  ifstream	helpFile(Config::getHelpFilename().c_str());
  if (!helpFile.is_open())
  {
    error = true;
    return StrList();
  }

  while (std::getline(helpFile, line))
  {
    std::string::size_type crlf = line.find_first_of("\r\n");
    if (crlf != std::string::npos)
    {
      line = line.substr(0, crlf);
    }

    if ((line.length() >= 2) && (line[0] == '.'))
    {
      if (!inTopic)
      {
	continue;
      }

      char code = line[1];
      std::string::size_type dot = line.find('.', 1);
      if (dot != std::string::npos)
      {
        std::string value = line.substr(dot + 1);

	if (code == 's')
	{
	  syntax.push_back(value);
	}

	if (code == 'd')
	{
	  description.push_back(value);
	}

	if (code == 'e')
	{
	  example.push_back(value);
	}

	if (code == 'f')
	{
	  if (value.find('c') != std::string::npos)
	  {
	    flags |= UF_CHANOP;
	  }
	  if (value.find('d') != std::string::npos)
	  {
	    flags |= UF_DLINE;
	  }
	  if (value.find('g') != std::string::npos)
	  {
	    flags |= UF_GLINE;
	  }
	  if (value.find('k') != std::string::npos)
	  {
	    flags |= UF_KLINE;
	  }
	  if (value.find('m') != std::string::npos)
	  {
	    flags |= UF_MASTER;
	  }
	  if (value.find('o') != std::string::npos)
	  {
	    flags |= UF_OPER;
	  }
	  if (value.find('r') != std::string::npos)
	  {
	    flags |= UF_REMOTE;
	  }
	  if (value.find('w') != std::string::npos)
	  {
	    flags |= UF_WALLOPS;
	  }
	}

	if (code == 'l')
	{
	  helpLink.push_back(value);
	}

	if (code == 't')
	{
	  subTopic.push_back(value);
	}
      }
    }
    else if ((line.length() >= 1) && (line[0] != '#'))
    {
      char *item;

      if (topics.size() > 0)
      {
	break;
      }

      StrSplit(topics, line, ".", true);
      inTopic = (topics.end() != std::find(topics.begin(), topics.end(),
	DownCase(topic)));
      if (!inTopic)
      {
	topics.clear();
      }
    }
  }
  helpFile.close();

  return getHelp();
}
