// Objected Oriented Monitor Bot
// OOMon by Timothy L. Jensen (aka Toast)
// February 1998

// $Id: main.cc,v 1.45 2003/05/25 17:17:50 toast Exp $

// Std C++ Headers
#include <iostream>
#include <fstream>
#include <string>
#include <list>

// Std C Headers
#include <sys/types.h>
#include <sys/time.h>
#include <signal.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <errno.h>

// OOMon Headers
#include "oomon.h"
#include "main.h"
#include "config.h"
#include "irc.h"
#include "services.h"
#include "log.h"
#include "proxy.h"
#include "util.h"
#include "dcclist.h"
#include "dcc.h"
#include "watch.h"
#include "vars.h"
#include "remote.h"
#include "botexcept.h"
#include "userdb.h"


#ifdef DEBUG
# define MAIN_DEBUG
#endif


#ifdef FORK_OOMON
static bool forkme = true;
#else
static bool forkme = false;
#endif

static std::string configFile = DEFAULT_CFGFILE;
static std::string pidFile = DEFAULT_PIDFILE;


void
SendAll(const std::string & message, const int flags,
  const WatchSet & watches, const DCC *exception)
{
  clients.sendAll(message, flags, watches, exception);
}


void
SendAll(const std::string & message, const int flags, Watch watch,
  const DCC *exception)
{
  WatchSet watches;
  watches.add(watch);
  clients.sendAll(message, flags, watches, exception);
}


void
gracefuldie(int sig)
{
  if (sig == SIGTERM)
  {
    ::SendAll("Caught SIGTERM -- OOMon terminating");
    Log::Write("Caught SIGTERM -- OOMon stopped");
    if (server.isConnected())
    {
      server.quit();
    }
  }
  else if (sig == SIGINT)
  {
    ::SendAll("Caught SIGINT -- OOMon terminating");
    Log::Write("Caught SIGINT -- OOMon stopped");
    if (server.isConnected())
    {
      server.quit("Caught SIGINT -- User pressed Ctrl+C?");
    }
  }
  else
  {
    if (server.isConnected())
    {
      server.quit("Abnormal termination!");
    }
    abort();
  }

  Log::Stop();

  if (userConfig)
  {
    delete userConfig;
  }

  exit(0);
}


void
reload()
{
  Config::Clear();
  Config::Load(configFile);
  Config::loadSettings();
  Log::Stop();
  Log::Start();
  Remote::Listen();
}


void
ReloadConfig(const std::string & From)
{
  ::SendAll("Reload CONFIG requested by " + From, UF_OPER);
  reload();
}


void
hangup(int sig)
{
  if (sig != SIGHUP)
    return;
  Log::Write("Caught SIGHUP -- Reloading config");
  ::SendAll("*** Caught SIGHUP.", UF_OPER);
  reload();
  signal(SIGHUP, hangup);
}


void
pipevec(int sig)
{
  if (sig == SIGPIPE)
  {
    cerr << "Caught SIGPIPE" << endl;
    signal(SIGPIPE, pipevec);
  }
}


bool
process()
{
  for (;;)
  {
    if (!server.isConnected() && !server.isConnecting())
    {
#ifdef MAIN_DEBUG
      cout << "Connecting to IRC server." << endl;
#endif
      if (!server.connect(Config::GetServerHostName(), Config::GetServerPort()))
      {
	server.reset();
      }
    }

    fd_set readfds, writefds;
    FD_ZERO(&readfds);
    FD_ZERO(&writefds);
    server.setFD(readfds, writefds);
    clients.setAllFD(readfds, writefds);
    Remote::SetFD(readfds);
    Proxy::setAllFD(readfds, writefds);

    struct timeval time_out;
    time_out.tv_sec = 5;
    time_out.tv_usec = 0;

    select(FD_SETSIZE, &readfds, &writefds, NULL, &time_out);

    try
    {
      if ((server.isConnected() || server.isConnecting()) &&
        !server.process(readfds, writefds))
      {
#ifdef MAIN_DEBUG
        cout << "Disconnected from IRC server." << endl;
#endif
        server.reset();
      }
    }
    catch (OOMon::timeout_error)
    {
#ifdef MAIN_DEBUG
      cout << "Connection from IRC server timed out." << endl;
#endif
      server.quit("Server inactive for " + IntToStr(server.getIdle()) +
        " seconds");
    }

    clients.processAll(readfds, writefds);

    Remote::Process(readfds);

    Proxy::processAll(readfds, writefds);

    if (server.isConnected())
    {
      services.check();
    }
  }
  return true;
}


// motd()
//
// Writes the Message Of The Day
//
void
motd(StrList & Output)
{
  ifstream	motdfile;
  char		line[MAX_BUFF];

  motdfile.open(Config::GetMOTD().c_str());
  if (!motdfile.fail())
  {
    Output.push_back("Message Of The Day:");
    while (motdfile.getline(line, MAX_BUFF-1)) {
      Output.push_back(std::string(line));
    }
    Output.push_back("End of MOTD");
  }
  else
  {
    Output.push_back("No MOTD");
  }
  motdfile.close();
}


void
printhelp(char *name)
{
  cout << "Usage:  " << name << " [-dh] [-c config_file] [-p pid_file]" << endl;
  cout << endl;
  cout << "Options:" << endl;
  cout << " -c config_file   load the specified config file" << endl;
  cout << " -d               do not fork into the background" << endl;
  cout << " -h               show this help" << endl;
  cout << " -p pid_file      write to the specified PID file" << endl;
}


int
parseargs(int argc, char **argv)
{
  int result = 0;
  int ch;

  while ((ch = getopt(argc, argv, "dhc:p:")) != -1)
  {
    switch (ch)
    {
      case 'c':
	configFile = optarg;
	break;
      case 'd':
        forkme = false;
	break;
      case 'p':
	pidFile = optarg;
	break;
      case 'h':
      case '?':
      default:
        printhelp(argv[0]);
	return -1;
	break;
    }
  }

  return result;
}


int
main(int argc, char **argv)
{
  if (parseargs(argc, argv))
  {
    return -1;
  }

  Config::Load(configFile);
  Config::loadSettings();

  try
  {
    userConfig = new UserDB(Config::getUserDBFile());
  }
  catch (OOMon::botdb_error & reason)
  {
    cerr << reason.what() <<  " (" << reason.why() << ")" << endl;
    exit(1);
  }

  if (forkme)
  {
    pid_t forkresult = fork();

    if (0 == forkresult)
    {
      // redirect output to /dev/null
      int fd = open("/dev/null", O_RDWR);
      dup2(fd, 1);	// stdout
      dup2(fd, 2);	// stderr
    }
    else
    {
      if (forkresult == -1)
      {
        cerr << "Unable to fork! Error " << IntToStr(errno) << endl;
        return 1;
      }
      return 0;
    }
  }

  signal(SIGSEGV, gracefuldie);
  signal(SIGBUS, gracefuldie);
  signal(SIGTERM, gracefuldie);
  signal(SIGINT, gracefuldie);
  signal(SIGHUP, hangup);
  signal(SIGPIPE, pipevec);

  ofstream PIDfile;
  PIDfile.open(pidFile.c_str());
  PIDfile << getpid();
  PIDfile.close();

  Log::Start();
  Log::Write("OOMon started");

  Remote::Listen();

  while (process());

  Log::Write("OOMon stopped");
  Log::Stop();

  if (userConfig)
  {
    delete userConfig;
  }

  return 0;
}
