// $Id: proxy.cc,v 1.36 2003/02/23 23:35:51 toast Exp $

// Std C++ headers
#include <iostream>
#include <fstream>
#include <string>
#include <list>

// Std C headers
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <fcntl.h>
#include <time.h>
#include <netdb.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <signal.h>
#include <pwd.h>

// OOMon headers
#include "oomon.h"
#include "proxy.h"
#include "socks4.h"
#include "wingate.h"
#include "botexcept.h"
#include "http.h"
#include "config.h"
#include "log.h"
#include "util.h"
#include "main.h"
#include "engine.h"
#include "irc.h"
#include "vars.h"


#ifdef DEBUG
# define PROXY_DEBUG
#endif

Proxy::ProxyList Proxy::items;
Proxy::Cache Proxy::safeHosts(4000);

const time_t Proxy::CACHE_EXPIRE = (60 * 60);	/* 1 hour */


Proxy::Proxy(const std::string & hostname, const std::string & nick,
  const std::string & userhost) : BotSock()
{
  this->hostname = hostname;
  this->nick = nick;
  this->userhost = userhost;
}


void
Proxy::detectedProxy()
{
  std::string notice = std::string("Open ") + this->getTypeName() +
    " proxy detected for " + this->getNick() + "!" + this->getUserhost() +
    " [" + this->getAddress() + ":" + IntToStr(this->getPort()) + "]";
  ::SendAll(notice, UF_OPER);
  Log::Write(notice);

  doAction(this->getNick(), this->getUserhost(),
    BotSock::inet_addr(this->getAddress()),
    vars[VAR_SCAN_PROXY_ACTION]->getAction(),
    vars[VAR_SCAN_PROXY_ACTION]->getInt(),
    vars[VAR_SCAN_PROXY_REASON]->getString(), false);
}


bool
Proxy::connect(const std::string & address, const BotSock::Port port)
{
  this->address = address;
  this->port = port;

  return this->BotSock::connect(address, port);
}


// connect(address, port)
//
// Attempts to connect to a proxy
//
// Returns true if no errors occurred
//
bool
Proxy::connect(Proxy *newProxy, const std::string & address,
  const BotSock::Port port)
{
  if (Proxy::isChecking(address, port, newProxy->getType()) ||
    Proxy::isSafe(address, port, newProxy->getType()))
  {
    delete newProxy;
    return false;
  }

#ifdef PROXY_DEBUG
  cout << "Creating proxy connection to address " << address << ":" << port <<
    endl;
#endif

  newProxy->bindTo(Config::getProxyVhost());
  newProxy->setTimeout(PROXY_IDLE_MAX);

  if (!newProxy->connect(address, port))
  {
    cerr << "Connect to proxy failed (other)" << endl;
    delete newProxy;
    return false;
  }

  Proxy::items.push_back(newProxy);

#ifdef PROXY_DEBUG
  cout << Proxy::items.size() << " proxies queued up for processing." << endl;
#endif

  return true;
}


// check(address, hostname, nick, userhost)
//
// Checks a client for proxies
//
// Returns true if no errors occurred
//
void
Proxy::check(const std::string & address, const std::string & hostname,
  const std::string & nick, const std::string & userhost)
{
  try
  {
    Proxy *newProxy;

    newProxy = new WinGate(hostname, nick, userhost);
    Proxy::connect(newProxy, address, 23);

    newProxy = new Socks4(hostname, nick, userhost);
    Proxy::connect(newProxy, address, 1080);

    newProxy = new Http(hostname, nick, userhost);
    Proxy::connect(newProxy, address, 80);

    newProxy = new Http(hostname, nick, userhost);
    Proxy::connect(newProxy, address, 1080);

    newProxy = new Http(hostname, nick, userhost);
    Proxy::connect(newProxy, address, 3128);

    newProxy = new Http(hostname, nick, userhost);
    Proxy::connect(newProxy, address, 8080);
  }
  catch (OOMon::errno_error & e)
  {
    cerr << "Connect to proxy failed (" << e.why() << ")" << endl;
  }
}


// SetFD(readset, writeset)
//
// Do this if you want to be able to receive data :P
//
void
Proxy::setAllFD(fd_set & readset, fd_set & writeset)
{
  for (ProxyList::iterator pos = Proxy::items.begin();
    pos != Proxy::items.end(); ++pos)
  {
    (*pos)->setFD(readset, writeset);
  }
}


// ProcessAll(readset, writeset)
//
// Processes any received data at each proxy connection.
//
void
Proxy::processAll(const fd_set & readset, const fd_set & writeset)
{
  for (ProxyList::iterator pos = Proxy::items.begin();
    pos != Proxy::items.end(); ++pos)
  {
    bool keepSocket;
    Proxy *tmp = (*pos);

    try
    {
      keepSocket = tmp->process(readset, writeset);
    }
    catch (OOMon::timeout_error)
    {
#ifdef PROXY_DEBUG
      cout << "Proxy connection to " << tmp->getAddress() << " timed out." <<
	endl;
#endif
      keepSocket = false;
    }

    if (!keepSocket)
    {
      --pos;
      Proxy::items.remove(tmp);

#ifdef PROXY_DEBUG
      cout << "Closing proxy connection to " << tmp->getAddress() << ":" <<
	tmp->getPort() << endl;
#endif
      delete tmp;

#ifdef PROXY_DEBUG
      cout << Proxy::items.size() << " proxies queued up for processing." <<
	endl;
#endif
    }
  }
}


bool
Proxy::isChecking(const std::string & address, const BotSock::Port port,
  const ProxyType type)
{
  for (ProxyList::iterator pos = Proxy::items.begin();
    pos != Proxy::items.end(); ++pos)
  {
    if (((*pos)->getAddress() == address) && ((*pos)->getPort() == port) &&
      ((*pos)->getType() == type))
    {
      return true;
    }
  }
  return false;
}


void
Proxy::addSafeHost(const std::string & address, const BotSock::Port port,
  const ProxyType type)
{
  const time_t now = time(NULL);
  const BotSock::Address ip = BotSock::inet_addr(address);
  int oldestIndex = 0;
  int oldestTime = now;

  if (INADDR_NONE != ip)
  {
    for (int index = 0; index < Proxy::safeHosts.size(); ++index)
    {
      // Is this an empty or expired entry?
      if (Proxy::safeHosts[index].isEmpty() ||
        (CACHE_EXPIRE <= (now - Proxy::safeHosts[index].checked)))
      {
        // Yes!  Use it.
        Proxy::safeHosts[index].ip = ip;
        Proxy::safeHosts[index].port = port;
        Proxy::safeHosts[index].type = type;
        Proxy::safeHosts[index].checked = now;

        return;
      }
      else
      {
        // No, but it might be the oldest entry
        if (oldestTime > Proxy::safeHosts[index].checked)
        {
          oldestIndex = index;
          oldestTime = Proxy::safeHosts[index].checked;
        }
      }
    }

    // We couldn't find an empty or expired cache entry, so just expire the
    // oldest entry we could find.
    Proxy::safeHosts[oldestIndex].ip = ip;
    Proxy::safeHosts[oldestIndex].port = port;
    Proxy::safeHosts[oldestIndex].type = type;
    Proxy::safeHosts[oldestIndex].checked = now;
  }
}


bool
Proxy::isSafe(const std::string & address, const BotSock::Port port,
  const ProxyType type)
{
  const time_t now = time(NULL);
  const BotSock::Address ip = BotSock::inet_addr(address);

  if (INADDR_NONE != ip)
  {
    for (Proxy::Cache::iterator pos = Proxy::safeHosts.begin();
      pos < Proxy::safeHosts.end(); ++pos)
    {
      if (!pos->isEmpty())
      {
        if ((pos->ip == ip) && (pos->port == port) && (pos->type == type) &&
          (CACHE_EXPIRE > (now - pos->checked)))
        {
          return true;
        }
      }
    }
  }
  return false;
}


void
Proxy::status(StrList & output)
{
  output.push_back("Proxy scanners: " + IntToStr(Proxy::items.size()));

  time_t now = time(NULL);
  int cacheCount = 0;
  for (Proxy::Cache::iterator pos = Proxy::safeHosts.begin();
    pos < Proxy::safeHosts.end(); ++pos)
  {
    if (!pos->isEmpty() && (CACHE_EXPIRE > (now - pos->checked)))
    {
      ++cacheCount;
    }
  }
  //output.push_back("Proxy cache: " + IntToStr(cacheCount) + "/" +
  //  IntToStr(Proxy::safeHosts.size()));
}

