// $Id: trap.cc,v 1.17 2003/05/01 18:21:52 toast Exp $

// Std C++ Headers
#include <fstream>
#include <string>

// OOMon Headers
#include "strtype"
#include "botexcept.h"
#include "trap.h"
#include "util.h"
#include "irc.h"
#include "log.h"
#include "main.h"
#include "config.h"
#include "vars.h"


list<Trap> TrapList::traps;


Trap::Trap(const TrapAction action, const long timeout,
  const std::string & nick, const std::string & userhost,
  const std::string & reason)
{
  this->action = action;
  this->timeout = timeout;
  this->nick = nick;
  this->userhost = userhost;
  this->reason = reason;
}

bool
Trap::matches(const std::string & nick, const std::string & userhost,
  const std::string & ip) const
{
  std::string::size_type at = userhost.find('@');

  std::string userip;
  if (std::string::npos != at)
  {
    userip = userhost.substr(0, at + 1) + ip;
  }

  return (MatchesMask(nick, this->nick, true) &&
    (MatchesMask(userhost, this->userhost) ||
    MatchesMask(userip, this->userhost)));
}


bool
Trap::exactMatch(const std::string & nick, const std::string & userhost) const
{
  return (Same(nick, this->nick) && Same(userhost, this->userhost));
}


TrapAction
Trap::getAction() const
{
  return this->action;
}


std::string
Trap::getString() const
{
  std::string text = TrapList::actionString(this->getAction()) + ": ";
  TrapAction action = this->getAction();
  switch (action)
  {
    case TRAP_ECHO:
      text += this->getMask();
      break;
    case TRAP_KILL:
    case TRAP_DLINE_IP:
    case TRAP_DLINE_NET:
      text += this->getMask() + " (" + this->getReason() + ")";
      break;
    case TRAP_KLINE:
    case TRAP_KLINE_HOST:
    case TRAP_KLINE_DOMAIN:
    case TRAP_KLINE_IP:
    case TRAP_KLINE_USERNET:
    case TRAP_KLINE_NET:
      if (this->getTimeout() > 0)
      {
        text += ULongToStr(this->getTimeout()) + " " + this->getMask() + " (" +
	  this->getReason() + ")";
      }
      else
      {
        text += this->getMask() + " (" + this->getReason() + ")";
      }
      break;
    default:
      text = "Unknown TRAP type!";
      break;
  }

  return text;
}

std::string
TrapList::actionString(const TrapAction & action)
{
  switch (action)
  {
    case TRAP_ECHO:
      return "ECHO";
    case TRAP_KILL:
      return "KILL";
    case TRAP_KLINE:
      return "KLINE";
    case TRAP_KLINE_HOST:
      return "KLINE_HOST";
    case TRAP_KLINE_DOMAIN:
      return "KLINE_DOMAIN";
    case TRAP_KLINE_IP:
      return "KLINE_IP";
    case TRAP_KLINE_USERNET:
      return "KLINE_USERNET";
    case TRAP_KLINE_NET:
      return "KLINE_NET";
    case TRAP_DLINE_IP:
      return "DLINE_IP";
    case TRAP_DLINE_NET:
      return "DLINE_NET";
  }
  return "";
}


TrapAction
TrapList::actionType(const std::string & text)
{
  if (Same("ECHO", text))
    return TRAP_ECHO;
  else if (Same("KILL", text))
    return TRAP_KILL;
  else if (Same("KLINE", text) || Same("KLINE_USERDOMAIN", text))
    return TRAP_KLINE;
  else if (Same("KLINE_HOST", text) || Same("KLINEHOST", text))
    return TRAP_KLINE_HOST;
  else if (Same("KLINE_DOMAIN", text))
    return TRAP_KLINE_DOMAIN;
  else if (Same("KLINE_IP", text))
    return TRAP_KLINE_IP;
  else if (Same("KLINE_USERNET", text))
    return TRAP_KLINE_USERNET;
  else if (Same("KLINE_NET", text))
    return TRAP_KLINE_NET;
  else if (Same("DLINE_IP", text) || Same("DLINE", text))
    return TRAP_DLINE_IP;
  else if (Same("DLINE_NET", text))
    return TRAP_DLINE_NET;

  throw OOMon::invalid_action(text);
}


long
Trap::getTimeout() const
{
  return this->timeout;
}


std::string
Trap::getMask() const
{
  return (this->nick + "!" + this->userhost);
}

std::string
Trap::getReason() const
{
  return this->reason;
}

void
TrapList::split(const std::string & mask, std::string & nick,
  std::string & userhost)
{
  std::string::size_type bang = mask.find('!');
  std::string::size_type at = mask.find('@');

  if (bang != std::string::npos)
  {
    nick = mask.substr(0, bang);
    userhost = mask.substr(bang + 1, std::string::npos);
  }
  else if (at != std::string::npos)
  {
    nick = "*";
    userhost = mask;
  }
  else
  {
    nick = mask;
    userhost = "*@*";
  }
}

Trap
TrapList::add(const TrapAction action, const long timeout,
  const std::string & mask, const std::string & reason)
{
  std::string nick;
  std::string userhost;

  split(mask, nick, userhost);

  remove(nick, userhost);

  Trap trap(action, timeout, nick, userhost, reason);

  traps.push_back(trap);

  return trap;
}


void
TrapList::cmd(StrList & output, std::string line, const std::string & handle)
{
  std::string flag = UpCase(FirstWord(line));
  std::string mask = FirstWord(line);
  long timeout = 0;
  bool modified = false;

  if (isNumeric(mask))
  {
    timeout = atol(mask.c_str());
    mask = FirstWord(line);
  }

  std::string reason = line;

  if (flag == "" || flag == "LIST")
  {
    TrapList::list(output);
  }
  else if (flag == "REMOVE")
  {
    if (TrapList::remove(mask))
    {
      modified = true;
      if (handle.size() > 0)
      {
        ::SendAll(handle + " removed trap: " + mask, UF_OPER,
	  WATCH_TRAPS, NULL);
        Log::Write(handle + " removed trap: " + mask);
      }
    }
    else
    {
      output.push_back(std::string("*** No trap exists for: ") + mask);
    }
  }
  else
  {
    try
    {
      TrapAction actionType = TrapList::actionType(flag);

      Trap node = TrapList::add(actionType, timeout, mask, reason);

      modified = true;

      if (handle.length() > 0)
      {
        std::string notice(handle + " added trap " + node.getString());
        ::SendAll(notice, UF_OPER, WATCH_TRAPS);
        Log::Write(notice);
      }
    }
    catch (OOMon::invalid_action & e)
    {
      output.push_back("*** Invalid TRAP action: " + e.what());
    }
  }

  if (modified && (handle.length() > 0) && vars[VAR_AUTO_SAVE]->getBool())
  {
    Config::saveSettings();
  }
}


bool
TrapList::remove(const std::string & mask)
{
  std::string nick;
  std::string userhost;

  split(mask, nick, userhost);

  return remove(nick, userhost);
}

bool
TrapList::remove(const std::string & nick, const std::string & userhost)
{
  bool found = false;
  std::list<Trap>::iterator pos = traps.begin();
  while (pos != traps.end())
  {
    if (pos->exactMatch(nick, userhost))
    {
      pos = traps.erase(pos);
      found = true;
    }
    else
    {
      ++pos;
    }
  }
  return found;
}


void
TrapList::match(const std::string & nick, const std::string & userhost,
  const std::string & ip)
{
  for (std::list<Trap>::iterator pos = traps.begin(); pos != traps.end(); ++pos)
  {
    if (!Config::IsOKHost(userhost, ip) && !Config::IsOper(userhost, ip) &&
      pos->matches(nick, userhost, ip))
    {
      std::string notice("*** Trapped: " + nick + "!" + userhost + " [" + ip +
	"]");

      BotSock::Address IP = BotSock::inet_addr(ip);

      Log::Write(notice);
      switch (pos->getAction())
      {
	case TRAP_ECHO:
	  ::SendAll(notice, UF_OPER, WATCH_TRAPS,
	    NULL);
	  break;
	case TRAP_KILL:
	  server.kill("Auto-Kill", nick, pos->getReason());
	  break;
	case TRAP_KLINE:
	  server.kline("Auto-Kline", pos->getTimeout(), klineMask(userhost),
	    pos->getReason());
	  break;
	case TRAP_KLINE_HOST:
	  {
	    std::string::size_type at = userhost.find('@');
	    if (at != std::string::npos)
	    {
	      std::string host = userhost.substr(at + 1, std::string::npos);
	      server.kline("Auto-Kline", pos->getTimeout(), "*@" + host,
	        pos->getReason());
	    }
	    else
	    {
	      cerr << "Userhost is missing @ symbol!" << endl;
	    }
	  }
	  break;
	case TRAP_KLINE_DOMAIN:
	  {
	    std::string::size_type at = userhost.find('@');
	    if (at != std::string::npos)
	    {
	      std::string domain = getDomain(userhost.substr(at + 1,
		std::string::npos), true);
	      server.kline("Auto-Kline", pos->getTimeout(), "*@" + domain,
	        pos->getReason());
	    }
	    else
	    {
	      cerr << "Userhost is missing @ symbol!" << endl;
	    }
	  }
	  break;
	case TRAP_KLINE_IP:
	  if (INADDR_NONE != IP)
	  {
	    server.kline("Auto-Kline", pos->getTimeout(), "*@" + ip,
	      pos->getReason());
	  }
	  else
	  {
	    cerr << "Missing or invalid IP address -- unable to KLINE!" << endl;
	  }
	  break;
	case TRAP_KLINE_USERNET:
	  {
	    std::string::size_type at = userhost.find('@');
	    if (at != std::string::npos)
	    {
	      std::string user = userhost.substr(0, at);
	      if ((user.length() > 1) && (user[0] == '~'))
	      {
		user = user.substr(1);
	      }
	      if (INADDR_NONE != IP)
	      {
	        server.kline("Auto-Kline", pos->getTimeout(), "*" + user + "@" +
		  classCMask(ip), pos->getReason());
	      }
	      else
	      {
	        cerr << "Missing or invalid IP address -- unable to KLINE!" <<
		  endl;
	      }
	    }
	    else
	    {
	      cerr << "Userhost is missing @ symbol!" << endl;
	    }
	  }
	  break;
	case TRAP_KLINE_NET:
	  if (INADDR_NONE != IP)
	  {
	    server.kline("Auto-Kline", pos->getTimeout(), "*@" + classCMask(ip),
	      pos->getReason());
	  }
	  else
	  {
	    cerr << "Missing or invalid IP address -- unable to KLINE!" << endl;
	  }
	  break;
	case TRAP_DLINE_IP:
	  if (INADDR_NONE != IP)
	  {
	    server.dline("Auto-Dline", ip, pos->getReason());
	  }
	  else
	  {
	    cerr << "Missing or invalid IP address -- unable to DLINE!" << endl;
	  }
	  break;
	case TRAP_DLINE_NET:
	  if (INADDR_NONE != IP)
	  {
	    server.dline("Auto-Dline", classCMask(ip), pos->getReason());
	  }
	  else
	  {
	    cerr << "Missing or invalid IP address -- unable to DLINE!" << endl;
	  }
	  break;
	default:
	  cerr << "Unknown Trap type!" << endl;
	  break;
      }
    }
  }
}

void
TrapList::list(StrList & output)
{
  int count = 0;

  for (std::list<Trap>::iterator pos = traps.begin(); pos != traps.end(); ++pos)
  {
    output.push_back(pos->getString());
    count++;
  }
  if (count > 0)
  {
    output.push_back("*** End of TRAP list.");
  }
  else
  {
    output.push_back("*** TRAP list is empty!");
  }
}


void
TrapList::save(ofstream & file)
{
  for (std::list<Trap>::iterator pos = TrapList::traps.begin();
    pos != TrapList::traps.end(); ++pos)
  {
    file << "TRAP " << TrapList::actionString(pos->getAction()) << " " <<
      pos->getTimeout() << " " << pos->getMask() << " " << pos->getReason() <<
      endl;
  }
}

